import ctypes # general doc about Structures: https://docs.python.org/3/library/ctypes.html#structures-and-unions

"""
Remark to the usage of ctypes:
All Python types except integers (int), strings (str), and bytes (byte) objects
have to be wrapped in their corresponding ctypes type, so that they can be
converted to the required C data type.

ctypes type     C type                  Python type
----------------------------------------------------------------
c_bool          _Bool                   bool (1)
c_char          char                    1-character bytes object
c_wchar         wchar_t                 1-character string
c_byte          char                    int
c_ubyte         unsigned char           int
c_short         short                   int
c_ushort        unsigned short          int
c_int           int                     int
c_uint          unsigned int            int
c_long          long                    int
c_ulong         unsigned long           int
c_longlong      __int64 or
                long long               int
c_ulonglong     unsigned __int64 or
                unsigned long long      int
c_size_t        size_t                  int
c_ssize_t       ssize_t or
                Py_ssize_t              int
c_float         float                   float
c_double        double                  float
c_longdouble    long double             float
c_char_p        char *
                (NUL terminated)        bytes object or None
c_wchar_p       wchar_t *
                (NUL terminated)        string or None
c_void_p        void *                  int or None

"""
# Reconstruct the proper structure of the variables, which can be extracted
# from the header file 'struct.h'.


class AcqStatus(ctypes.Structure):
    """ Create a structured Data type with ctypes where the dll can write into.

    This object handles and retrieves the acquisition status data from the
    Fastcomtec.

    int started;                // acquisition status: 1 if running, 0 else
    double runtime;             // running time in seconds
    double totalsum;            // total events
    double roisum;              // events within ROI
    double roirate;             // acquired ROI-events per second
    double nettosum;            // ROI sum with background subtracted
    double sweeps;              // Number of sweeps
    double stevents;            // Start Events
    unsigned long maxval;       // Maximum value in spectrum
    """
    def __init__(self):
        super().__init__()
        self.started    : ctypes.c_int  
        self.runtime    : ctypes.c_double
        self.totalsum   : ctypes.c_double
        self.roisum     : ctypes.c_double
        self.roirate    : ctypes.c_double
        self.ofls       : ctypes.c_double  
        self.sweeps     : ctypes.c_double
        self.stevents   : ctypes.c_double
        self.maxval     : ctypes.c_ulong 

    _fields_ = [('started', ctypes.c_int),
                ('runtime', ctypes.c_double),
                ('totalsum', ctypes.c_double),
                ('roisum', ctypes.c_double),
                ('roirate', ctypes.c_double),
                ('ofls', ctypes.c_double),
                ('sweeps', ctypes.c_double),
                ('stevents', ctypes.c_double),
                ('maxval', ctypes.c_ulong), ]
    
    def __repr__(self) -> str:
        return  'started : ' + str(self.started ) + '\n' \
                'runtime : ' + str(self.runtime ) + '\n' \
                'totalsum: ' + str(self.totalsum) + '\n' \
                'roisum  : ' + str(self.roisum  ) + '\n' \
                'roirate : ' + str(self.roirate ) + '\n' \
                'ofls    : ' + str(self.ofls    ) + '\n' \
                'sweeps  : ' + str(self.sweeps  ) + '\n' \
                'stevents: ' + str(self.stevents) + '\n' \
                'maxval  : ' + str(self.maxval  )

class AcqData(ctypes.Structure):
    """ Create a structured Data type with ctypes where the dll can write into.

    This object handles and retrieves the acquisition data of the Fastcomtec.
    """
    def __init__(self):
        super().__init__()
        self.s0     : ctypes._Pointer 
        self.region : ctypes._Pointer 
        self.comment: ctypes.c_char_p 
        self.cnt    : ctypes._Pointer 
        self.hs0    : ctypes.c_int    
        self.hrg    : ctypes.c_int    
        self.hcm    : ctypes.c_int    
        self.hct    : ctypes.c_int    

    _fields_ = [('s0', ctypes.POINTER(ctypes.c_ulong)),
                ('region', ctypes.POINTER(ctypes.c_ulong)),
                ('comment', ctypes.c_char_p),
                ('cnt', ctypes.POINTER(ctypes.c_double)),
                ('hs0', ctypes.c_int),
                ('hrg', ctypes.c_int),
                ('hcm', ctypes.c_int),
                ('hct', ctypes.c_int), ]
    
    def __repr__(self) -> str:
        return  's0     : ' + str(self.s0     ) + '\n' \
                'region : ' + str(self.region ) + '\n' \
                'comment: ' + str(self.comment) + '\n' \
                'cnt    : ' + str(self.cnt    ) + '\n' \
                'hs0    : ' + str(self.hs0    ) + '\n' \
                'hrg    : ' + str(self.hrg    ) + '\n' \
                'hcm    : ' + str(self.hcm    ) + '\n' \
                'hct    : ' + str(self.hct    )

class AcqSettings(ctypes.Structure):
    def __init__(self):
        super().__init__()
        self.range          : ctypes.c_long     
        self.cftfak         : ctypes.c_long     
        self.roimin         : ctypes.c_long     
        self.roimax         : ctypes.c_long     
        self.nregions       : ctypes.c_long     
        self.caluse         : ctypes.c_long     
        self.calpoints      : ctypes.c_long     
        self.param          : ctypes.c_long     
        self.offset         : ctypes.c_long     
        self.xdim           : ctypes.c_long     
        self.bitshift       : ctypes.c_ulong    
        self.active         : ctypes.c_long     
        self.eventpreset    : ctypes.c_double   
        self.dummy1         : ctypes.c_double   
        self.dummy2         : ctypes.c_double   
        self.dummy3         : ctypes.c_double   

    _fields_ = [('range',       ctypes.c_long),
                ('cftfak',      ctypes.c_long),
                ('roimin',      ctypes.c_long),
                ('roimax',      ctypes.c_long),
                ('nregions',    ctypes.c_long),
                ('caluse',      ctypes.c_long),
                ('calpoints',   ctypes.c_long),
                ('param',       ctypes.c_long),
                ('offset',      ctypes.c_long),
                ('xdim',        ctypes.c_long),
                ('bitshift',    ctypes.c_ulong),
                ('active',      ctypes.c_long),
                ('eventpreset', ctypes.c_double),
                ('dummy1',      ctypes.c_double),
                ('dummy2',      ctypes.c_double),
                ('dummy3',      ctypes.c_double), ]
    
    def __repr__(self) -> str:
        return  'range      : ' + str(self.range      ) + '\n' \
                'cftfak     : ' + str(self.cftfak     ) + '\n' \
                'roimin     : ' + str(self.roimin     ) + '\n' \
                'roimax     : ' + str(self.roimax     ) + '\n' \
                'nregions   : ' + str(self.nregions   ) + '\n' \
                'caluse     : ' + str(self.caluse     ) + '\n' \
                'calpoints  : ' + str(self.calpoints  ) + '\n' \
                'param      : ' + str(self.param      ) + '\n' \
                'offset     : ' + str(self.offset     ) + '\n' \
                'xdim       : ' + str(self.xdim       ) + '\n' \
                'bitshift   : ' + str(self.bitshift   ) + '\n' \
                'active     : ' + str(self.active     ) + '\n' \
                'eventpreset: ' + str(self.eventpreset) + '\n' \
                'dummy1     : ' + str(self.dummy1     ) + '\n' \
                'dummy2     : ' + str(self.dummy2     ) + '\n' \
                'dummy3     : ' + str(self.dummy3     )

class BoardSetting(ctypes.Structure):
    def __init__(self):
        super().__init__()
        self.sweepmode : ctypes.c_long
        self.prena     : ctypes.c_long
        self.cycles    : ctypes.c_long
        self.sequences : ctypes.c_long
        self.syncout   : ctypes.c_long
        self.digio     : ctypes.c_long
        self.digval    : ctypes.c_long
        self.dac0      : ctypes.c_long
        self.dac1      : ctypes.c_long
        self.dac2      : ctypes.c_long
        self.dac3      : ctypes.c_long
        self.dac4      : ctypes.c_long
        self.dac5      : ctypes.c_long
        self.fdac      : ctypes.c_int
        self.tagbits   : ctypes.c_int
        self.extclk    : ctypes.c_int
        self.maxchan   : ctypes.c_long
        self.serno     : ctypes.c_long
        self.ddruse    : ctypes.c_long
        self.active    : ctypes.c_long
        self.holdafter : ctypes.c_double
        self.swpreset  : ctypes.c_double
        self.fstchan   : ctypes.c_double
        self.timepreset: ctypes.c_double

    _fields_ = [('sweepmode',   ctypes.c_long),
                ('prena',       ctypes.c_long),
                ('cycles',      ctypes.c_long),
                ('sequences',   ctypes.c_long),
                ('syncout',     ctypes.c_long),
                ('digio',       ctypes.c_long),
                ('digval',      ctypes.c_long),
                ('dac0',        ctypes.c_long),
                ('dac1',        ctypes.c_long),
                ('dac2',        ctypes.c_long),
                ('dac3',        ctypes.c_long),
                ('dac4',        ctypes.c_long),
                ('dac5',        ctypes.c_long),
                ('fdac',        ctypes.c_int),
                ('tagbits',     ctypes.c_int),
                ('extclk',      ctypes.c_int),
                ('maxchan',     ctypes.c_long),
                ('serno',       ctypes.c_long),
                ('ddruse',      ctypes.c_long),
                ('active',      ctypes.c_long),
                ('holdafter',   ctypes.c_double),
                ('swpreset',    ctypes.c_double),
                ('fstchan',     ctypes.c_double),
                ('timepreset',  ctypes.c_double),]
    
    def __repr__(self) -> str:
        return  'sweepmode : ' + str(self.sweepmode ) + '\n' \
                'prena     : ' + str(self.prena     ) + '\n' \
                'cycles    : ' + str(self.cycles    ) + '\n' \
                'sequences : ' + str(self.sequences ) + '\n' \
                'syncout   : ' + str(self.syncout   ) + '\n' \
                'digio     : ' + str(self.digio     ) + '\n' \
                'digval    : ' + str(self.digval    ) + '\n' \
                'dac0      : ' + str(self.dac0      ) + '\n' \
                'dac1      : ' + str(self.dac1      ) + '\n' \
                'dac2      : ' + str(self.dac2      ) + '\n' \
                'dac3      : ' + str(self.dac3      ) + '\n' \
                'dac4      : ' + str(self.dac4      ) + '\n' \
                'dac5      : ' + str(self.dac5      ) + '\n' \
                'fdac      : ' + str(self.fdac      ) + '\n' \
                'tagbits   : ' + str(self.tagbits   ) + '\n' \
                'extclk    : ' + str(self.extclk    ) + '\n' \
                'maxchan   : ' + str(self.maxchan   ) + '\n' \
                'serno     : ' + str(self.serno     ) + '\n' \
                'ddruse    : ' + str(self.ddruse    ) + '\n' \
                'active    : ' + str(self.active    ) + '\n' \
                'holdafter : ' + str(self.holdafter ) + '\n' \
                'swpreset  : ' + str(self.swpreset  ) + '\n' \
                'fstchan   : ' + str(self.fstchan   ) + '\n' \
                'timepreset: ' + str(self.timepreset) 
